import unittest
from datetime import datetime, timezone
from ipaddress import IPv4Network, IPv6Network
from pathlib import Path

import pytest
import responses
import yaml
from dateutil.tz import tzoffset
from responses import _recorder

import whoisit


BASE_DIR = Path(__file__).resolve().parent
RESPONSES = BASE_DIR / "responses"
UTC = timezone.utc


class SyncPublicInterfaceTestCase(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        whoisit.clear_bootstrapping()

    def test_version(self):
        self.assertRegex(whoisit.version, r"\d+\.\d+\.\d+")

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'boostrap.yaml')
    def test_bootstrap_interface(self):
        responses._add_from_file(RESPONSES / "boostrap.yaml")
        whoisit.bootstrap()
        self.assertTrue(whoisit.is_bootstrapped())

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'asn-1.yaml')
    def test_asn_interface_1(self):
        responses._add_from_file(RESPONSES / "asn-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.asn(13335)
        self.assertEqual(resp["type"], "autnum")
        self.assertEqual(resp["handle"], "AS13335")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/autnum/13335")
        self.assertEqual(resp["handle"], "AS13335")
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["asn_range"], [13335, 13335])
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(resp["copyright_notice"], "Copyright 1997-2024, American Registry for Internet Numbers, Ltd.")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["registration_date"], datetime(2010, 7, 14, 18, 35, 57, tzinfo=tzoffset(None, -14400)))
        self.assertTrue(isinstance(resp["entities"]["registrant"], list))
        self.assertTrue(len(resp["entities"]["registrant"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["abuse"], list))
        self.assertTrue(len(resp["entities"]["abuse"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["noc"], list))
        self.assertTrue(len(resp["entities"]["noc"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["technical"], list))
        self.assertTrue(len(resp["entities"]["technical"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'domain-1.yaml')
    def test_domain_interface_1(self):
        responses._add_from_file(RESPONSES / "domain-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.domain("google.com")
        self.assertEqual(resp["type"], "domain")
        self.assertEqual(resp["name"], "GOOGLE.COM")
        self.assertEqual(resp["handle"], "2138514_DOMAIN_COM-VRSN")
        self.assertEqual(resp["rir"], "")
        self.assertEqual(resp["registration_date"], datetime(1997, 9, 15, 4, 0, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.verisign.com/com/v1/domain/GOOGLE.COM")
        self.assertEqual(
            resp["terms_of_service_url"],
            "https://www.verisign.com/domain-names/registration-data-access-protocol/terms-service/index.xhtml",
        )
        self.assertEqual(resp["whois_server"], "")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(resp["dnssec"], False)
        self.assertEqual(resp["nameservers"], ["NS1.GOOGLE.COM", "NS2.GOOGLE.COM", "NS3.GOOGLE.COM", "NS4.GOOGLE.COM"])
        self.assertEqual(
            resp["status"],
            [
                "client delete prohibited",
                "client transfer prohibited",
                "client update prohibited",
                "server delete prohibited",
                "server transfer prohibited",
                "server update prohibited",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["registrar"], list))
        self.assertTrue(len(resp["entities"]["registrar"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'ip-v4-1.yaml')
    def test_ip_interface_v4_1(self):
        responses._add_from_file(RESPONSES / "ip-v4-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.ip("1.1.1.1")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "APNIC-LABS")
        self.assertEqual(resp["handle"], "1.1.1.0 - 1.1.1.255")
        self.assertEqual(resp["network"], IPv4Network("1.1.1.0/24"))
        self.assertEqual(resp["rir"], "apnic")
        self.assertEqual(resp["registration_date"], datetime(2011, 8, 10, 23, 12, 35, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.apnic.net/ip/1.1.1.0/24")
        self.assertEqual(resp["terms_of_service_url"], "http://www.apnic.net/db/dbcopyright.html")
        self.assertEqual(resp["whois_server"], "whois.apnic.net")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(
            resp["description"],
            [
                "APNIC and Cloudflare DNS Resolver project",
                "Routed globally by AS13335/Cloudflare",
                "Research prefix for APNIC Labs",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'ip-v6-1.yaml')
    def test_ip_interface_v6_1(self):
        responses._add_from_file(RESPONSES / "ip-v6-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.ip("2606:4700:4700::1111")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["handle"], "NET6-2606-4700-1")
        self.assertEqual(resp["network"], IPv6Network("2606:4700::/32"))
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2011, 11, 1, 15, 59, 58, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/ip/2606:4700::")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["All Cloudflare abuse reporting can be done via https://www.cloudflare.com/abuse"]
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'ip-v4-cidr-1.yaml')
    def test_ip_interface_v4_cidr_1(self):
        responses._add_from_file(RESPONSES / "ip-v4-cidr-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.ip("1.1.1.0/24")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "APNIC-LABS")
        self.assertEqual(resp["handle"], "1.1.1.0 - 1.1.1.255")
        self.assertEqual(resp["rir"], "apnic")
        self.assertEqual(resp["registration_date"], datetime(2011, 8, 10, 23, 12, 35, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.apnic.net/ip/1.1.1.0/24")
        self.assertEqual(resp["terms_of_service_url"], "http://www.apnic.net/db/dbcopyright.html")
        self.assertEqual(resp["whois_server"], "whois.apnic.net")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(resp["assignment_type"], "assigned portable")
        self.assertEqual(resp["country"], "AU")
        self.assertEqual(resp["ip_version"], 4)
        self.assertEqual(resp["network"], IPv4Network("1.1.1.0/24"))
        self.assertEqual(
            resp["description"],
            [
                "APNIC and Cloudflare DNS Resolver project",
                "Routed globally by AS13335/Cloudflare",
                "Research prefix for APNIC Labs",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'ip-v6-cidr-1.yaml')
    def test_ip_interface_v6_cidr_1(self):
        responses._add_from_file(RESPONSES / "ip-v6-cidr-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.ip("2606:4700::/32")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["handle"], "NET6-2606-4700-1")
        self.assertEqual(resp["network"], IPv6Network("2606:4700::/32"))
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2011, 11, 1, 15, 59, 58, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/ip/2606:4700::")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["All Cloudflare abuse reporting can be done via https://www.cloudflare.com/abuse"]
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'entity-1.yaml')
    def test_entity_interface_1(self):
        responses._add_from_file(RESPONSES / "entity-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.entity("GOVI", rir="arin")
        self.assertEqual(resp["type"], "entity")
        self.assertEqual(resp["name"], "Govital Internet Inc.")
        self.assertEqual(resp["handle"], "GOVI")
        self.assertEqual(resp["parent_handle"], "")
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2001, 5, 8, 0, 0, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["expiration_date"], None)
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/entity/GOVI")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["http://www.govital.net\r", "Standard NOC hours are 10am to 6pm EST M-F"]
        )
        self.assertTrue(isinstance(resp["entities"]["technical"], list))
        self.assertTrue(len(resp["entities"]["technical"]) > 0)


#############################################
################### Async ###################
#############################################


def load_sync_responses_to_httpx_mock(file, httpx_mock):
    with open(file) as fh:
        data = yaml.safe_load(fh)

    for item in data["responses"]:
        response = item["response"]
        httpx_mock.add_response(
            url=response["url"],
            method=response["method"],
            status_code=response["status"],
            text=response["body"],
            headers=response["headers"],
        )


@pytest.fixture(scope="function")
def mock_httpx(request, httpx_mock):
    request.cls.httpx_mock = httpx_mock


class AsyncPublicInterfaceTestCase(unittest.IsolatedAsyncioTestCase):
    maxDiff = None

    def setUp(self):
        whoisit.clear_bootstrapping()

    def test_version(self):
        self.assertRegex(whoisit.version, r"\d+\.\d+\.\d+")

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_bootstrap_interface(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "boostrap.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        self.assertTrue(whoisit.is_bootstrapped())

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_asn_interface_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "asn-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.asn_async(13335)
        self.assertEqual(resp["type"], "autnum")
        self.assertEqual(resp["handle"], "AS13335")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/autnum/13335")
        self.assertEqual(resp["handle"], "AS13335")
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["asn_range"], [13335, 13335])
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(resp["copyright_notice"], "Copyright 1997-2024, American Registry for Internet Numbers, Ltd.")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["registration_date"], datetime(2010, 7, 14, 18, 35, 57, tzinfo=tzoffset(None, -14400)))
        self.assertTrue(isinstance(resp["entities"]["registrant"], list))
        self.assertTrue(len(resp["entities"]["registrant"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["abuse"], list))
        self.assertTrue(len(resp["entities"]["abuse"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["noc"], list))
        self.assertTrue(len(resp["entities"]["noc"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["technical"], list))
        self.assertTrue(len(resp["entities"]["technical"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_domain_interface_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "domain-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.domain_async("google.com")
        self.assertEqual(resp["type"], "domain")
        self.assertEqual(resp["name"], "GOOGLE.COM")
        self.assertEqual(resp["handle"], "2138514_DOMAIN_COM-VRSN")
        self.assertEqual(resp["rir"], "")
        self.assertEqual(resp["registration_date"], datetime(1997, 9, 15, 4, 0, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.verisign.com/com/v1/domain/GOOGLE.COM")
        self.assertEqual(
            resp["terms_of_service_url"],
            "https://www.verisign.com/domain-names/registration-data-access-protocol/terms-service/index.xhtml",
        )
        self.assertEqual(resp["whois_server"], "")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(resp["dnssec"], False)
        self.assertEqual(resp["nameservers"], ["NS1.GOOGLE.COM", "NS2.GOOGLE.COM", "NS3.GOOGLE.COM", "NS4.GOOGLE.COM"])
        self.assertEqual(
            resp["status"],
            [
                "client delete prohibited",
                "client transfer prohibited",
                "client update prohibited",
                "server delete prohibited",
                "server transfer prohibited",
                "server update prohibited",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["registrar"], list))
        self.assertTrue(len(resp["entities"]["registrar"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_ip_interface_v4_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "ip-v4-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.ip_async("1.1.1.1")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "APNIC-LABS")
        self.assertEqual(resp["handle"], "1.1.1.0 - 1.1.1.255")
        self.assertEqual(resp["network"], IPv4Network("1.1.1.0/24"))
        self.assertEqual(resp["rir"], "apnic")
        self.assertEqual(resp["registration_date"], datetime(2011, 8, 10, 23, 12, 35, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.apnic.net/ip/1.1.1.0/24")
        self.assertEqual(resp["terms_of_service_url"], "http://www.apnic.net/db/dbcopyright.html")
        self.assertEqual(resp["whois_server"], "whois.apnic.net")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(
            resp["description"],
            [
                "APNIC and Cloudflare DNS Resolver project",
                "Routed globally by AS13335/Cloudflare",
                "Research prefix for APNIC Labs",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_ip_interface_v6_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "ip-v6-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.ip_async("2606:4700:4700::1111")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["handle"], "NET6-2606-4700-1")
        self.assertEqual(resp["network"], IPv6Network("2606:4700::/32"))
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2011, 11, 1, 15, 59, 58, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/ip/2606:4700::")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["All Cloudflare abuse reporting can be done via https://www.cloudflare.com/abuse"]
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_ip_interface_v4_cidr_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "ip-v4-cidr-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.ip_async("1.1.1.0/24")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "APNIC-LABS")
        self.assertEqual(resp["handle"], "1.1.1.0 - 1.1.1.255")
        self.assertEqual(resp["rir"], "apnic")
        self.assertEqual(resp["registration_date"], datetime(2011, 8, 10, 23, 12, 35, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.apnic.net/ip/1.1.1.0/24")
        self.assertEqual(resp["terms_of_service_url"], "http://www.apnic.net/db/dbcopyright.html")
        self.assertEqual(resp["whois_server"], "whois.apnic.net")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(resp["assignment_type"], "assigned portable")
        self.assertEqual(resp["country"], "AU")
        self.assertEqual(resp["ip_version"], 4)
        self.assertEqual(resp["network"], IPv4Network("1.1.1.0/24"))
        self.assertEqual(
            resp["description"],
            [
                "APNIC and Cloudflare DNS Resolver project",
                "Routed globally by AS13335/Cloudflare",
                "Research prefix for APNIC Labs",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_ip_interface_v6_cidr_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "ip-v6-cidr-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.ip_async("2606:4700::/32")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["handle"], "NET6-2606-4700-1")
        self.assertEqual(resp["network"], IPv6Network("2606:4700::/32"))
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2011, 11, 1, 15, 59, 58, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/ip/2606:4700::")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["All Cloudflare abuse reporting can be done via https://www.cloudflare.com/abuse"]
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_entity_interface_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "entity-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.entity_async("GOVI", rir="arin")
        self.assertEqual(resp["type"], "entity")
        self.assertEqual(resp["name"], "Govital Internet Inc.")
        self.assertEqual(resp["handle"], "GOVI")
        self.assertEqual(resp["parent_handle"], "")
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2001, 5, 8, 0, 0, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["expiration_date"], None)
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/entity/GOVI")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["http://www.govital.net\r", "Standard NOC hours are 10am to 6pm EST M-F"]
        )
        self.assertTrue(isinstance(resp["entities"]["technical"], list))
        self.assertTrue(len(resp["entities"]["technical"]) > 0)



from urllib3.util import retry

try:
    from urllib3.util import create_urllib3_context
except ImportError:
    from urllib3.util.ssl_ import create_urllib3_context

import httpx
import requests
from urllib3.poolmanager import PoolManager

from .errors import QueryError, UnsupportedError
from .logger import get_logger
from .version import version


log = get_logger('utils')
user_agent = 'whoisit/{version}'
insecure_ssl_ciphers = 'ALL:@SECLEVEL=1'
http_timeout = 10                    # Maximum time in seconds to allow for an HTTP request
http_retry_statuses = [429]          # HTTP status codes to trigger a retry wih backoff
http_max_retries = 3                 # Maximum number of HTTP requests to retry before failing
http_pool_connections = 10           # Maximum number of HTTP pooled connections
http_pool_maxsize = 10               # Maximum HTTP pool connection size
async_http_max_connections = 100     # Maximum number of HTTP connections allowed for async client
async_max_keepalive_connections = 20 # Allow the connection pool to maintain keep-alive connections below this point
_default_session = {'secure': None, 'insecure': False}


def get_session_or_async_client(session_or_async_client=None, allow_insecure_ssl=False, is_async=False):
    """
        Creates and caches the default sessions, one for secure (default) SSL
        and one for SSL with an insecure cipher suite.
    """
    global _default_session
    if session_or_async_client:
        if allow_insecure_ssl:
            if not _default_session['insecure']:
                _default_session['insecure'] = session_or_async_client
            return _default_session['insecure']
        else:
            if not _default_session['secure']:
                _default_session['secure'] = session_or_async_client
            return _default_session['secure']
    else:
        if allow_insecure_ssl:
            if is_async:
                _default_session['insecure'] = create_async_client(allow_insecure_ssl=allow_insecure_ssl)
            else:
                _default_session['insecure'] = create_session(allow_insecure_ssl=allow_insecure_ssl)
            return _default_session['insecure']
        else:
            if is_async:
                _default_session['secure'] = create_async_client(allow_insecure_ssl=allow_insecure_ssl)
            else:
                _default_session['secure'] = create_session(allow_insecure_ssl=allow_insecure_ssl)
            return _default_session['secure']


def get_session(session=None, allow_insecure_ssl=False) -> requests.Session:
    return get_session_or_async_client(session, allow_insecure_ssl, is_async=False)


def get_async_client(client=None, allow_insecure_ssl=False) -> httpx.AsyncClient:
    return get_session_or_async_client(client, allow_insecure_ssl, is_async=True)


class InsecureSSLAdapter(requests.adapters.HTTPAdapter):
    """
        Custom adapter to permit insecure SSL connections.
    """

    def init_poolmanager(self, connections, maxsize, block=False):
        insecure_ssl_ciphersuite = create_urllib3_context(ciphers=insecure_ssl_ciphers)
        self.poolmanager = PoolManager(ssl_context=insecure_ssl_ciphersuite)


def create_session(allow_insecure_ssl=False):
    session = requests.session()
    session_retry = retry.Retry(total=http_max_retries,
                                status_forcelist=http_retry_statuses,
                                backoff_factor=1)
    whoisit_adapter = requests.adapters.HTTPAdapter(
        max_retries=session_retry,
        pool_connections=http_pool_connections,
        pool_maxsize=http_pool_maxsize)
    session.mount('https://', whoisit_adapter)
    if allow_insecure_ssl:
        session.mount('https://', InsecureSSLAdapter())
    return session


def create_async_client(allow_insecure_ssl=False):
    limits = httpx.Limits(max_connections=async_http_max_connections, max_keepalive_connections=async_max_keepalive_connections)
    retries = httpx.AsyncHTTPTransport(retries=http_max_retries, limits=limits)
    headers = {"User-Agent": user_agent.format(version=version)}
    verify = not allow_insecure_ssl
    client = httpx.AsyncClient(transport=retries, verify=verify, headers=headers, follow_redirects=True)
    return client


def http_request(session, url, method='GET', headers=None, data=None, *args, **kwargs):
    """
        Simple wrapper over requests. Allows for optionally downgrading SSL
        ciphers if required.
    """
    headers = headers or {}
    data = data or {}
    methods = ('GET',)
    if method not in methods:
        raise UnsupportedError(f'HTTP methods supported are: {methods}, got: {method}')
    headers['User-Agent'] = user_agent.format(version=version)
    log.debug(f'Making HTTP {method} request to {url}')
    try:
        if 'timeout' not in kwargs:
            kwargs['timeout'] = http_timeout
        return session.request(method, url, headers=headers, data=data, *args, **kwargs)
    except Exception as e:
        raise QueryError(f'Failed to make a {method} request to {url}: {e}') from e


async def http_request_async(client: httpx.AsyncClient, url, method='GET', headers=None, data=None, *args, **kwargs):
    """
        Simple wrapper over httpx.
    """
    headers = headers or {}
    data = data or {}
    methods = ('GET',)
    if method not in methods:
        raise UnsupportedError(f'HTTP methods supported are: {methods}, got: {method}')

    log.debug(f'Making async HTTP {method} request to {url}')
    try:
        if 'timeout' not in kwargs:
            kwargs['timeout'] = http_timeout
        return await client.request(method, url, headers=headers, data=data, *args, **kwargs)
    except Exception as e:
        raise QueryError(f'Failed to make a {method} request to {url}: {e}') from e


def is_subnet_of(network_a, network_b):
    a_len = network_a.prefixlen
    b_len = network_b.prefixlen
    return a_len >= b_len and network_a.supernet(a_len - b_len) == network_b


default_chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-'


def contains_only_chars(s, chars=default_chars):
    for c in s:
        if c not in chars:
            return False
    return True


def recursive_merge(d1, d2):
    for k, v in d2.items():
        if k in d1 and isinstance(d1[k], dict) and isinstance(v, dict):
            recursive_merge(d1[k], v)
        elif v:
            d1[k] = v



class WhoisItError(Exception):
    """
        Parent Exception for all whoisit raised exceptions.
    """


class BootstrapError(WhoisItError):
    """
        Raised when there are any issues with bootstrapping.
    """


class QueryError(WhoisItError):
    """
        Raised when there are any issues with queries.
    """

    def __init__(self, message, status_code=0, response=''):
        super().__init__(message)
        self.status_code = status_code
        self.response = response


class UnsupportedError(WhoisItError):
    """
        Raised when a feature in a query is unsupported.
    """


class ParseError(WhoisItError):
    """
        Raised when failing to parse response data.
    """


class ResourceDoesNotExist(QueryError):
    """
        Raised when querying a resource which doesn't exist.
    """


class ResourceAccessDeniedError(QueryError):
    """
        Raised when querying a resource returns an access denied response.
    """


class RateLimitedError(QueryError):
    """
        Raised when querying a resource and getting a rate limited response.
    """


class RemoteServerError(QueryError):
    """
        Raised when querying a resource and getting a remote server error response.
    """



import os
import sys
import logging


debug_env_var = str(os.getenv('DEBUG', '')).strip().lower()
is_debugging = debug_env_var in ('yes', 'y', '1', 'true')
default_level = logging.DEBUG if is_debugging else logging.INFO


def get_logger(name, level=default_level):
    log = logging.getLogger(name)
    log.setLevel(level)
    ch = logging.StreamHandler()
    ch.setLevel(level)
    fmt = logging.Formatter('%(asctime)s %(name)s [%(levelname)s] %(message)s')
    ch.setFormatter(fmt)
    log.addHandler(ch)
    return log
